import re
import time

import pytest

from freiner.strategies import MovingWindowRateLimiter


def test_pluggable_storage_no_moving_window():
    class MyStorage:
        def incr(self, key: str, expiry: int, elastic_expiry: bool = False) -> int:
            return 1

        def get(self, key: str) -> int:
            return 0

        def get_expiry(self, key: str) -> int:
            return int(time.time())

    storage = MyStorage()

    errmsg = re.escape("MovingWindowRateLimiting is not implemented for storage of type MyStorage")
    with pytest.raises(TypeError, match=errmsg):
        # Ignore the type error here because that's exactly what we're testing for.
        MovingWindowRateLimiter(storage)  # type: ignore


def test_pluggable_storage_moving_window():
    class MyStorage:
        def incr(self, key: str, expiry: int, elastic_expiry: bool = False) -> int:
            return 1

        def get(self, key: str) -> int:
            return 0

        def get_expiry(self, key: str) -> int:
            return int(time.time())

        def check(self) -> bool:
            return True

        def clear(self, key: str):
            pass

        def reset(self):
            pass

        def acquire_entry(self, *a, **k) -> bool:
            return True

        def get_moving_window(self, *a, **k):
            return time.time(), 1

    storage = MyStorage()
    strategy = MovingWindowRateLimiter(storage)
    assert strategy.storage is storage



import time
from typing import List, Tuple, Union
from urllib.parse import urlparse

from freiner.errors import FreinerConfigurationError


try:
    import pymemcache

    HAS_MEMCACHED = True
except ImportError:  # pragma: no cover
    HAS_MEMCACHED = False


MemcachedClient = Union[pymemcache.Client, pymemcache.PooledClient, pymemcache.HashClient]


class MemcachedStorage:
    """
    Rate limit storage with memcached as backend.

    Depends on the `pymemcache` library.
    """

    MAX_CAS_RETRIES = 10

    def __init__(self, client: MemcachedClient):
        self._client: MemcachedClient = client

    @classmethod
    def from_uri(cls, uri: str, **options) -> "MemcachedStorage":
        """
        :param str uri: memcached location of the form
         `memcached://host:port,host:port`, `memcached:///run/path/to/sock`
        :param options: all remaining keyword arguments are passed
         directly to the constructor of :class:`pymemcache.client.base.Client`
        :raise FreinerConfigurationError: when `pymemcache` dependency is not available
        """

        if not HAS_MEMCACHED:
            raise FreinerConfigurationError("Dependency 'pymemcache' is not available.")

        parsed_uri = urlparse(uri)
        hosts: List[Union[Tuple[str, int], str]] = []
        for loc in parsed_uri.netloc.strip().split(","):
            if not loc:
                continue

            host, port = loc.split(":")
            hosts.append((host, int(port)))
        else:
            # filesystem path to UDS
            if parsed_uri.path and not parsed_uri.netloc and not parsed_uri.port:
                hosts = [parsed_uri.path]

        if not hosts:
            raise FreinerConfigurationError(f"No Memcached hosts parsed from URI: {uri}")

        if len(hosts) > 1:
            client = pymemcache.HashClient(hosts, **options)
        else:
            client = pymemcache.Client(*hosts, **options)
        return cls(client)

    def get(self, key: str) -> int:
        """
        :param str key: the key to get the counter value for
        """
        return int(self._client.get(key) or 0)

    def clear(self, key: str):
        """
        :param str key: the key to clear rate limits for
        """
        self._client.delete(key)

    def incr(self, key: str, expiry: int, elastic_expiry: bool = False) -> int:
        """
        increments the counter for a given rate limit key

        :param str key: the key to increment
        :param int expiry: amount in seconds for the key to expire in
        :param bool elastic_expiry: whether to keep extending the rate limit
         window every hit.
        """

        if not self._client.add(key, 1, expiry, noreply=False):
            if not elastic_expiry:
                return self._client.incr(key, 1) or 1

            value, cas = self._client.gets(key)
            retry = 0

            while (
                not self._client.cas(key, int(value or 0) + 1, cas, expiry)
                and retry < self.MAX_CAS_RETRIES
            ):
                value, cas = self._client.gets(key)
                retry += 1

            self._client.set(key + "/expires", expiry + time.time(), expire=expiry, noreply=False)
            return int(value or 0) + 1

        self._client.set(key + "/expires", expiry + time.time(), expire=expiry, noreply=False)
        return 1

    def get_expiry(self, key: str) -> int:
        """
        :param str key: the key to get the expiry for
        """
        return int(float(self._client.get(key + "/expires") or time.time()))

    def check(self) -> bool:
        """
        check if storage is healthy
        """
        try:
            self._client.get("freiner-check")
            return True
        except:  # noqa
            return False

    def reset(self):
        """
        Not implemented for Memcached.
        """
        pass



from typing import Optional
from urllib.parse import urlparse

from freiner.errors import FreinerConfigurationError

from .redis import RedisStorage


try:
    import redis.sentinel

    HAS_REDIS = True
except ImportError:  # pragma: no cover
    HAS_REDIS = False


class RedisSentinelStorage(RedisStorage):
    """
    Rate limit storage with redis sentinel as backend.

    Depends on `redis` library.
    """

    def __init__(self, sentinel: redis.sentinel.Sentinel, service_name: str):
        self._sentinel: redis.sentinel.Sentinel = sentinel
        self._service_name: str = service_name

        self._sentinel_master: redis.Redis = self._sentinel.master_for(self._service_name)
        self._sentinel_slave: redis.Redis = self._sentinel.slave_for(self._service_name)

        super().__init__(self._sentinel_master)

    @classmethod
    def from_uri(
        cls, uri: str, service_name: Optional[str] = None, **options
    ) -> "RedisSentinelStorage":
        """
        :param str uri: url of the form
         `redis+sentinel://host:port,host:port/service_name`
        :param Optional[str] service_name: sentinel service name
         (if not provided in `uri`)
        :param options: all remaining keyword arguments are passed
         directly to the constructor of :class:`redis.sentinel.Sentinel`
        :raise FreinerConfigurationError: when the redis library is not available
         or if the redis master host cannot be pinged.
        """

        if not HAS_REDIS:
            raise FreinerConfigurationError("Dependency 'redis' is not available.")

        parsed_uri = urlparse(uri)
        sentinel_configuration = []

        password = None
        if parsed_uri.password:
            password = parsed_uri.password

        for loc in parsed_uri.netloc[parsed_uri.netloc.find("@") + 1 :].split(","):
            host, port = loc.split(":")
            sentinel_configuration.append((host, int(port)))

        service_name = parsed_uri.path.replace("/", "") if parsed_uri.path else service_name
        if service_name is None:
            raise FreinerConfigurationError("'service_name' not provided")

        options.setdefault("socket_timeout", 0.2)

        sentinel = redis.sentinel.Sentinel(sentinel_configuration, password=password, **options)
        return cls(sentinel, service_name)

    def get(self, key: str) -> int:
        """
        :param str key: the key to get the counter value for
        """
        return self._get(key, self._sentinel_slave)

    def get_expiry(self, key: str) -> int:
        """
        :param str key: the key to get the expiry for
        """
        return self._get_expiry(key, self._sentinel_slave)

    def check(self) -> bool:
        """
        check if storage is healthy
        """
        return self._check(self._sentinel_slave)



from typing import Tuple

from typing_extensions import Protocol, runtime_checkable

from .memory import MemoryStorage


@runtime_checkable
class Storage(Protocol):
    def incr(self, key: str, expiry: int, elastic_expiry: bool = False) -> int:
        """
        increments the counter for a given rate limit key

        :param str key: the key to increment
        :param int expiry: amount in seconds for the key to expire in
        :param bool elastic_expiry: whether to keep extending the rate limit
         window every hit.
        """

    def get(self, key: str) -> int:
        """
        :param str key: the key to get the counter value for
        """

    def get_expiry(self, key: str) -> int:
        """
        :param str key: the key to get the expiry for
        """

    def check(self) -> bool:
        """
        check if storage is healthy
        """

    def clear(self, key: str):
        """
        resets the rate limit key
        :param str key: the key to clear rate limits for
        """

    def reset(self):
        """
        reset storage to clear limits
        """


@runtime_checkable
class MovingWindowStorage(Storage, Protocol):
    def acquire_entry(self, key: str, limit: int, expiry: int, no_add: bool = False) -> bool:
        """
        :param str key: rate limit key to acquire an entry in
        :param int limit: amount of entries allowed
        :param int expiry: expiry of the entry
        :param bool no_add: if False an entry is not actually acquired
         but instead serves as a 'check'
        :rtype: bool
        """

    def get_moving_window(self, key: str, limit: int, expiry: int) -> Tuple[int, int]:
        """
        returns the starting point and the number of entries in the moving window

        :param str key: rate limit key
        :param int limit: amount of entries allowed
        :param int expiry: expiry of entry
        :return: (start of window, number of acquired entries)
        """


__all__ = [
    "Storage",
    "MovingWindowStorage",
    "MemoryStorage",
]



import os
import sys

from invoke import task, util


pty = util.isatty(sys.stdout) and util.isatty(sys.stderr)


@task
def reformat(c):
    c.run("isort freiner tests setup.py tasks.py", pty=pty)
    c.run("black freiner tests setup.py tasks.py doc/source/conf.py", pty=pty)


@task
def lint(c):
    c.run("flake8 --show-source --statistics --max-line-length 100 freiner tests", pty=pty)
    c.run("check-manifest", pty=pty)


@task
def test(c, onefile=""):
    pytest_args = ["pytest", "--cov=freiner", "--cov-branch", "--cov-report=term"]
    if os.environ.get("CI", "false") == "true":
        pytest_args.append("--cov-report=xml")
        _pty = False
    else:
        pytest_args.append("--cov-report=html")
        _pty = pty

    if onefile:
        pytest_args.append(onefile)

    c.run("docker-compose down --remove-orphans --volumes", pty=_pty)
    c.run("docker-compose up -d", pty=_pty)
    try:
        c.run(" ".join(pytest_args), pty=_pty)
    finally:
        c.run("docker-compose down --remove-orphans --volumes", pty=_pty)

        c.run("rm .docker/memcached/freiner.memcached.sock", pty=_pty, warn=True)
        c.run("rm .docker/redis/freiner.redis.sock", pty=_pty, warn=True)

@task
def type_check(c):
    c.run("mypy freiner tests", pty=pty)



"""
rate limiting strategies
"""

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING

from .limits import RateLimitItem
from .storage import MovingWindowStorage, Storage


class RateLimiter(metaclass=ABCMeta):
    def __init__(self, storage: Storage):
        self.storage: Storage = storage

    @abstractmethod
    def hit(self, item: RateLimitItem, *identifiers) -> bool:
        """
        creates a hit on the rate limit and returns True if successful.

        :param item: a :class:`RateLimitItem` instance
        :param identifiers: variable list of strings to uniquely identify the limit
        :return: True/False
        """
        raise NotImplementedError

    @abstractmethod
    def test(self, item: RateLimitItem, *identifiers) -> bool:
        """
        checks  the rate limit and returns True if it is not
        currently exceeded.

        :param item: a :class:`RateLimitItem` instance
        :param identifiers: variable list of strings to uniquely identify the limit
        :return: True/False
        """
        raise NotImplementedError

    @abstractmethod
    def get_window_stats(self, item: RateLimitItem, *identifiers):
        """
        returns the number of requests remaining and reset of this limit.

        :param item: a :class:`RateLimitItem` instance
        :param identifiers: variable list of strings to uniquely identify the limit
        :return: tuple (reset time (int), remaining (int))
        """
        raise NotImplementedError

    def clear(self, item: RateLimitItem, *identifiers):
        return self.storage.clear(item.key_for(*identifiers))


class MovingWindowRateLimiter(RateLimiter):
    """
    Reference: :ref:`moving-window`
    """

    def __init__(self, storage: MovingWindowStorage):
        if not isinstance(storage, MovingWindowStorage):
            msg = f"MovingWindowRateLimiting is not implemented for storage of type {storage.__class__.__name__}"
            raise TypeError(msg)

        super().__init__(storage)
        if TYPE_CHECKING:
            self.storage: MovingWindowStorage = storage

    def hit(self, item: RateLimitItem, *identifiers) -> bool:
        """
        creates a hit on the rate limit and returns True if successful.

        :param item: a :class:`RateLimitItem` instance
        :param identifiers: variable list of strings to uniquely identify the limit
        :return: True/False
        """
        return self.storage.acquire_entry(
            item.key_for(*identifiers), item.amount, item.get_expiry()
        )

    def test(self, item: RateLimitItem, *identifiers) -> bool:
        """
        checks  the rate limit and returns True if it is not
        currently exceeded.

        :param item: a :class:`RateLimitItem` instance
        :param identifiers: variable list of strings to uniquely identify the limit
        :return: True/False
        """

        _, acquired_count = self.storage.get_moving_window(
            item.key_for(*identifiers), item.amount, item.get_expiry()
        )

        return acquired_count < item.amount

    def get_window_stats(self, item: RateLimitItem, *identifiers):
        """
        returns the number of requests remaining within this limit.

        :param item: a :class:`RateLimitItem` instance
        :param identifiers: variable list of strings to uniquely identify the limit
        :return: tuple (reset time (int), remaining (int))
        """
        window_start, window_items = self.storage.get_moving_window(
            item.key_for(*identifiers), item.amount, item.get_expiry()
        )
        reset = window_start + item.get_expiry()
        return reset, item.amount - window_items


class FixedWindowRateLimiter(RateLimiter):
    """
    Reference: :ref:`fixed-window`
    """

    def hit(self, item: RateLimitItem, *identifiers) -> bool:
        """
        creates a hit on the rate limit and returns True if successful.

        :param item: a :class:`RateLimitItem` instance
        :param identifiers: variable list of strings to uniquely identify the limit
        :return: True/False
        """
        return self.storage.incr(item.key_for(*identifiers), item.get_expiry()) <= item.amount

    def test(self, item: RateLimitItem, *identifiers) -> bool:
        """
        checks  the rate limit and returns True if it is not
        currently exceeded.

        :param item: a :class:`RateLimitItem` instance
        :param identifiers: variable list of strings to uniquely identify the limit
        :return: True/False
        """
        return self.storage.get(item.key_for(*identifiers)) < item.amount

    def get_window_stats(self, item: RateLimitItem, *identifiers):
        """
        returns the number of requests remaining and reset of this limit.

        :param item: a :class:`RateLimitItem` instance
        :param identifiers: variable list of strings to uniquely identify the limit
        :return: tuple (reset time (int), remaining (int))
        """
        remaining = max(0, item.amount - self.storage.get(item.key_for(*identifiers)))
        reset = self.storage.get_expiry(item.key_for(*identifiers))
        return reset, remaining


class FixedWindowElasticExpiryRateLimiter(FixedWindowRateLimiter):
    """
    Reference: :ref:`fixed-window-elastic`
    """

    def hit(self, item: RateLimitItem, *identifiers) -> bool:
        """
        creates a hit on the rate limit and returns True if successful.

        :param item: a :class:`RateLimitItem` instance
        :param identifiers: variable list of strings to uniquely identify the limit
        :return: True/False
        """

        counter = self.storage.incr(
            item.key_for(*identifiers), item.get_expiry(), elastic_expiry=True
        )
        return counter <= item.amount



import re
from typing import Sequence, Type, cast

from .limits import GRANULARITIES, RateLimitItem


# TODO: Remove {1}
SEPARATORS = re.compile(r"[,;|]{1}")
# TODO: Mark second group as non-capturing, adjust parse_many to remove _ variable
SINGLE_EXPR = re.compile(
    r"""
    \s*([0-9]+)
    \s*(/|\s*per\s*)
    \s*([0-9]+)
    *\s*(hour|minute|second|day|month|year)s?\s*""",
    re.IGNORECASE | re.VERBOSE,
)
EXPR = re.compile(
    r"^{SINGLE}(:?{SEPARATORS}{SINGLE})*$".format(
        SINGLE=SINGLE_EXPR.pattern, SEPARATORS=SEPARATORS.pattern
    ),
    re.IGNORECASE | re.VERBOSE,
)


def parse_many(limit_string: str) -> Sequence[RateLimitItem]:
    """
    parses rate limits in string notation containing multiple rate limits
    (e.g. '1/second; 5/minute')

    :param string limit_string: rate limit string using :ref:`ratelimit-string`
    :raise ValueError: if the string notation is invalid.
    :return: a list of :class:`RateLimitItem` instances.
    """
    if not isinstance(limit_string, str):
        raise TypeError("Invalid rate limit string supplied.")

    if not EXPR.match(limit_string):
        raise ValueError(f"Failed to parse rate limit string: {limit_string}")

    limits = []
    for limit in SEPARATORS.split(limit_string):
        # This cast is fine because we already verified that it will match
        # in the EXPR.match check above.
        limit_match = cast(re.Match, SINGLE_EXPR.match(limit))
        amount, _, multiples, granularity_string = limit_match.groups()
        granularity = granularity_from_string(granularity_string)
        limits.append(granularity(amount, multiples))

    return tuple(limits)


def parse(limit_string: str) -> RateLimitItem:
    """
    parses a single rate limit in string notation
    (e.g. '1/second' or '1 per second'

    :param string limit_string: rate limit string using :ref:`ratelimit-string`
    :raise ValueError: if the string notation is invalid.
    :return: an instance of :class:`RateLimitItem`
    """
    return parse_many(limit_string)[0]


def granularity_from_string(granularity_string: str) -> Type[RateLimitItem]:
    """
    :param granularity_string:
    :return: a subclass of :class:`RateLimitItem`
    :raise ValueError:
    """
    for granularity in GRANULARITIES.values():
        if granularity.check_granularity_string(granularity_string):
            return granularity

    raise ValueError(f"No granularity matched for: {granularity_string}")



import threading
import time
from collections import Counter
from typing import Tuple


class LockableEntry(threading._RLock):
    __slots__ = ["atime", "expiry"]

    def __init__(self, expiry):
        self.atime = time.time()
        self.expiry = self.atime + expiry
        super(LockableEntry, self).__init__()


class MemoryStorage:
    """
    rate limit storage using :class:`collections.Counter`
    as an in memory storage for fixed and elastic window strategies,
    and a simple list to implement moving window strategy.
    """

    def __init__(self):
        self.storage = Counter()
        self.expirations = {}
        self.events = {}
        self.timer = threading.Timer(0.01, self.__expire_events)
        self.timer.start()

    def __expire_events(self):
        for key in list(self.events.keys()):
            for event in list(self.events[key]):
                with event:
                    if event.expiry <= time.time() and event in self.events[key]:
                        self.events[key].remove(event)

        for key in list(self.expirations.keys()):
            if self.expirations[key] <= time.time():
                self.storage.pop(key, None)
                self.expirations.pop(key, None)

    def __schedule_expiry(self):
        if not self.timer.is_alive():
            self.timer = threading.Timer(0.01, self.__expire_events)
            self.timer.start()

    def incr(self, key: str, expiry: int, elastic_expiry: bool = False) -> int:
        """
        increments the counter for a given rate limit key

        :param str key: the key to increment
        :param int expiry: amount in seconds for the key to expire in
        :param bool elastic_expiry: whether to keep extending the rate limit
         window every hit.
        """
        self.get(key)
        self.__schedule_expiry()
        self.storage[key] += 1
        if elastic_expiry or self.storage[key] == 1:
            self.expirations[key] = time.time() + expiry
        return self.storage.get(key, 0)

    def get(self, key: str) -> int:
        """
        :param str key: the key to get the counter value for
        """
        if self.expirations.get(key, 0) <= time.time():
            self.storage.pop(key, None)
            self.expirations.pop(key, None)
        return self.storage.get(key, 0)

    def clear(self, key: str):
        """
        :param str key: the key to clear rate limits for
        """
        self.storage.pop(key, None)
        self.expirations.pop(key, None)
        self.events.pop(key, None)

    def acquire_entry(self, key: str, limit: int, expiry: int, no_add: bool = False) -> bool:
        """
        :param str key: rate limit key to acquire an entry in
        :param int limit: amount of entries allowed
        :param int expiry: expiry of the entry
        :param bool no_add: if False an entry is not actually acquired
         but instead serves as a 'check'
        :rtype: bool
        """
        self.events.setdefault(key, [])
        self.__schedule_expiry()
        timestamp = time.time()
        try:
            entry = self.events[key][limit - 1]
        except IndexError:
            entry = None
        if entry and entry.atime >= timestamp - expiry:
            return False
        else:
            if not no_add:
                self.events[key].insert(0, LockableEntry(expiry))
            return True

    def get_expiry(self, key: str) -> int:
        """
        :param str key: the key to get the expiry for
        """
        return int(self.expirations.get(key, -1))

    def get_num_acquired(self, key: str, expiry: int) -> int:
        """
        returns the number of entries already acquired

        :param str key: rate limit key to acquire an entry in
        :param int expiry: expiry of the entry
        """
        timestamp = time.time()
        if self.events.get(key):
            return len([k for k in self.events[key] if k.atime >= timestamp - expiry])
        else:
            return 0

    def get_moving_window(self, key: str, limit: int, expiry: int) -> Tuple[int, int]:
        """
        returns the starting point and the number of entries in the moving window

        :param str key: rate limit key
        :param int limit: amount of entries allowed
        :param int expiry: expiry of entry
        :return: (start of window, number of acquired entries)
        """
        timestamp = time.time()
        acquired = self.get_num_acquired(key, expiry)
        for item in self.events.get(key, []):
            if item.atime >= timestamp - expiry:
                return int(item.atime), acquired
        return int(timestamp), acquired

    def check(self) -> bool:
        """
        check if storage is healthy
        """
        return True

    def reset(self):
        self.storage.clear()
        self.expirations.clear()
        self.events.clear()



import time

import pytest

from freiner.limits import RateLimitItemPerMinute, RateLimitItemPerSecond
from freiner.storage import MemoryStorage
from freiner.strategies import (
    FixedWindowElasticExpiryRateLimiter,
    FixedWindowRateLimiter,
    MovingWindowRateLimiter,
)

from tests import freeze_time


@pytest.fixture
def storage() -> MemoryStorage:
    return MemoryStorage()


def test_fixed_window(storage: MemoryStorage):
    limiter = FixedWindowRateLimiter(storage)
    with freeze_time() as frozen_datetime:
        limit = RateLimitItemPerSecond(10, 2)
        start = int(time.time())

        assert all([limiter.hit(limit) for _ in range(0, 10)]) is True

        frozen_datetime.tick(1)
        assert limiter.hit(limit) is False
        assert limiter.get_window_stats(limit)[1] == 0
        assert limiter.get_window_stats(limit)[0] == start + 2

        frozen_datetime.tick(1)
        assert limiter.get_window_stats(limit)[1] == 10
        assert limiter.hit(limit) is True


def test_fixed_window_with_elastic_expiry(storage: MemoryStorage):
    limiter = FixedWindowElasticExpiryRateLimiter(storage)
    with freeze_time() as frozen_datetime:
        limit = RateLimitItemPerSecond(10, 2)
        start = int(time.time())

        assert all([limiter.hit(limit) for _ in range(0, 10)]) is True

        frozen_datetime.tick(1)
        assert limiter.hit(limit) is False
        assert limiter.get_window_stats(limit)[1] == 0
        # three extensions to the expiry
        assert limiter.get_window_stats(limit)[0] == start + 3

        frozen_datetime.tick(1)
        assert limiter.hit(limit) is False

        frozen_datetime.tick(3)
        start = int(time.time())
        assert limiter.hit(limit) is True
        assert limiter.get_window_stats(limit)[1] == 9
        assert limiter.get_window_stats(limit)[0] == start + 2


def test_moving_window_in_memory(storage: MemoryStorage):
    limiter = MovingWindowRateLimiter(storage)
    with freeze_time() as frozen_datetime:
        limit = RateLimitItemPerMinute(10)
        for i in range(0, 5):
            assert limiter.hit(limit) is True
            assert limiter.hit(limit) is True
            assert limiter.get_window_stats(limit)[1] == 10 - ((i + 1) * 2)
            frozen_datetime.tick(10)

        assert limiter.get_window_stats(limit)[1] == 0
        assert limiter.hit(limit) is False

        frozen_datetime.tick(20)
        assert limiter.get_window_stats(limit)[1] == 2
        assert limiter.get_window_stats(limit)[0] == int(time.time() + 30)

        frozen_datetime.tick(31)
        assert limiter.get_window_stats(limit)[1] == 10


def test_test_fixed_window(storage: MemoryStorage):
    limiter = FixedWindowRateLimiter(storage)
    with freeze_time():
        limit = RateLimitItemPerSecond(2, 1)

        assert limiter.test(limit) is True
        assert limiter.hit(limit) is True
        assert limiter.test(limit) is True
        assert limiter.hit(limit) is True
        assert limiter.test(limit) is False
        assert limiter.hit(limit) is False


def test_test_moving_window(storage: MemoryStorage):
    limiter = MovingWindowRateLimiter(storage)
    with freeze_time():
        limit = RateLimitItemPerSecond(2, 1)

        assert limiter.test(limit) is True
        assert limiter.hit(limit) is True
        assert limiter.test(limit) is True
        assert limiter.hit(limit) is True
        assert limiter.test(limit) is False
        assert limiter.hit(limit) is False



import os
import sys

sys.path.insert(0, os.path.abspath("../../"))
import freiner

on_rtd = os.environ.get("READTHEDOCS", None) == "True"

if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme

    html_theme = "sphinx_rtd_theme"
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

autodoc_default_options = {
    "members": True,
    "show-inheritance": True,
}

autoclass_content = "both"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
]

source_suffix = ".rst"
master_doc = "index"
project = "Freiner"
copyright = "2021, Matthew Gamble"

version = release = freiner.__version__
exclude_patterns = []
pygments_style = "sphinx"
htmlhelp_basename = "freinerdoc"

latex_documents = [
    ("index", "limits.tex", "Freiner Documentation", "Matthew Gamble", "manual"),
]
man_pages = [("index", "freiner", "Freiner Documentation", ["Matthew Gamble"], 1)]

texinfo_documents = [
    (
        "index",
        "freiner",
        "Freiner Documentation",
        "Matthew Gamble",
        "freiner",
        "One line description of project.",
        "Miscellaneous",
    ),
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/", None),
    "redis": ("https://redis-py.readthedocs.io/", None),
    "rediscluster": ("https://redis-py-cluster.readthedocs.io/", None),
    "pymemcache": ("https://pymemcache.readthedocs.io/", None),
}



from .limits import (
    RateLimitItem,
    RateLimitItemPerDay,
    RateLimitItemPerHour,
    RateLimitItemPerMinute,
    RateLimitItemPerMonth,
    RateLimitItemPerSecond,
    RateLimitItemPerYear,
)
from .util import parse, parse_many


__version__ = "2.0.0"

__all__ = [
    "RateLimitItem",
    "RateLimitItemPerYear",
    "RateLimitItemPerMonth",
    "RateLimitItemPerDay",
    "RateLimitItemPerHour",
    "RateLimitItemPerMinute",
    "RateLimitItemPerSecond",
    "parse",
    "parse_many",
]
